#version 140

/////////////////////////////////////////////////////////////////////////////IN
//
// The texture sampler is needed for texture mapping.
//
uniform sampler2D TextureMap;

// Material parameters
uniform vec4 MaterialEmission;  // = vec4();  // Ecm
uniform vec4 MaterialAbmient;   // = vec4(0.5, 0.0, 0.0, 1.0); // Acm
uniform vec4 MaterialDiffuse;   // = vec4(0.4, 0.4, 0.5, 1.0); // Dcm
uniform vec4 MaterialSpecular;  // = vec4(0.8, 0.8, 0.0, 1.0); // Scm
uniform float MaterialShininess;// = vec4(20.0); // Srm
uniform float MaterialAlpha;    // = [0 ... transparent; 1 ... opaque]

// Scene
vec4 SceneColor; // Derived. Ecm + Acm * Acs

// Light parameters
uniform vec3 LightPosition;
uniform vec3 HalfVector;

uniform vec4 LightColor;
uniform vec4 LightAmbient;
uniform vec4 LightDiffuse;
uniform vec4 LightSpecular;

in vec3 Normal;
in vec4 TexCoord;

////////////////////////////////////////////////////////////////////////////OUT
out vec4 FragColor;

///////////////////////////////////////////////////////////////HELPER FUNCTIONS
float ndotl(vec3 normal, vec3 lightPosition)
{
    return max(0.0, dot(normalize(normal), lightPosition));
}
float ndoth(vec3 normal, vec3 halfVector)
{
    return max(0.0, dot(normal, halfVector));
}
float powerFactor(float nDotH, float materialShininess)
{
    return (nDotL == 0.0) ? 0.0 : pow(nDotH, materialShininess)
}
///////////////////////////////////////////////////////////////////////////MAIN
void main()
{
    vec3 n = normalize(Normal);

    //
    // Determine the scene color (Ecm + Acm * Acs).
    //
    vec4 SceneColor = 
    //
    // nDotL ... 
    //
    float nDotL = ndotl(Normal, LightPosition);
    float nDotH = ndoth(Normal, HalfVector);
    float power = powerFactor(nDotH, MaterialShininess);
    
    //
    // Determine the color of the mesh without the texture color.
    //
    vec4 ambient  = MaterialAbmient;
    vec4 diffuse  = MaterialDiffuse  * nDotL;
    vec4 specular = MaterialSpecular * power;
    vec4 color    = SceneColor       + ambient + diffuse + specular;
    
    //
    // The color of the fragment is the color of the mesh times the color of
    // the texel at the position of the texture coordinate.
    //
    FragColor = color * texture2D(TextureMap, TexCoord.st);
    FragColor.a = MaterialAlpha;
}
