#version 140

// This matrix is used to transform the eye coordinates to clip coordinates
// (can be changed once per frame).
uniform mat4 ProjectionMatrix;

// This matrix is used to transform the object coordinates in order to produce
// eye coordinates (can be changed once per frame).
uniform mat4 ViewMatrix;

// The light position is used to calculate the reflection of light on the
// material (can be changed once per frame).
uniform vec3 LightPosition;

// Position vector of the vertex.
in vec3 Position;

// The multi-texture coordinate of the first texture.
in vec4 MultiTexCoord0;

// Normal vector of the vertex.
inout vec3 Normal;

out vec4 TexCoord;

void main()
{
    // Normalize the light position and the vertex normal (i.e. the normal
    // vector in the vertex position).
    Normal = normalize(LightPosition * Normal);

    // Transform the vertex position and map it to clipping coordinates.
    gl_Position = ProjectionMatrix * ViewMatrix * vec4(Position, 1.0);

    // Set the vertex texture coordinate for texture unit 0 by copying the
    // texture coordinate specified in the application.
    // gl_TexCoord ... is varying, i.e. it will be used in the fragment shader
    //      to access the interpolated texture coordinate.
    TexCoord = MultiTexCoord0;
}
