#version 140

#define NUM_PYRAMID_LEVELS 17 // k=2
//#define NUM_ENV_SWATHS 32

#define NUM_ENV_SWATHS 8


uniform sampler2DArray heightField; //Bilinear Access
uniform sampler2D test;

uniform sampler2D vis4;
uniform sampler2D vis8;
uniform sampler2D vis12;
uniform sampler2D vis16;

uniform float RedLightEnv[16];
uniform float GreenLightEnv[16];
uniform float BlueLightEnv[16];

uniform int offset;

uniform float phi;

in vec2 ex_TexCoord;
out vec4 ex_Color;

// B-spline weights
const vec4 w_knot = vec4(1.0/6.0, 2.0/3.0, 1.0/6.0, 0.0);
const vec4 w_mid = vec4(1.0/48.0, 23.0/48.0, 23.0/48.0, 1.0/48.0);


const float multiResBase = pow(2.0, 1.0/2.0);
const float HALF_PI = 1.5707963;
const float ONE_OVER_HALF_PI = 0.31830988618379067153776752674503;
//const float ONE_OVER_HALF_PI = 0.63661978322714409245807365347117;
//const float M_PI = 3.1415926;
const float M_PI = 3.1415926535897932384626433832795;

// Stuff for ZH rotation
//float zhemi_canon[4] = float[4](1.0/M_PI * 0.886227,1.0/M_PI *  1.02333,1.0/M_PI *  0.495416,1.0/M_PI *  0.0);
float zhemi_canon[4] = float[4](0.886227, 1.02333, 0.495416, 0.0);

#define CONSTANT(x) float(x)

float fastAtan(float x)
{
	if (abs(x) < 1.0)
	{
		return x/(1.0 + 0.28*x*x); 
	}
	else
	{
		if (x >= 0.0)
			return HALF_PI - x/(x*x + 0.28);
		else
			return -HALF_PI - x/(x*x + 0.28);
	}
}

//
// Rotate SH vector in Y around z axis by angle whose cosine and sine
// are given by the first two arguments, respectively.  The result
// is added into the output vector Yr.
//
void SH_rotatez_4(in float c, in float s, in float Y[16], out float Yr[16])
{
    float cm = c;
    float sm = s;
    float newcm;

    // m=0
    Yr[ 0] = Y[ 0];
    Yr[ 2] = Y[ 2];
    Yr[ 6] = Y[ 6];
    Yr[12] = Y[12];

    // m=1
    Yr[ 1] = cm*Y[ 1] + sm*Y[ 3];
    Yr[ 3] = cm*Y[ 3] - sm*Y[ 1];
    Yr[ 5] = cm*Y[ 5] + sm*Y[ 7];
    Yr[ 7] = cm*Y[ 7] - sm*Y[ 5];
    Yr[11] = cm*Y[11] + sm*Y[13];
    Yr[13] = cm*Y[13] - sm*Y[11];

    newcm = c*cm - s*sm;
    sm = c*sm + s*cm;
    cm = newcm;

    // m=2
    Yr[ 4] = cm*Y[ 4] + sm*Y[ 8];
    Yr[ 8] = cm*Y[ 8] - sm*Y[ 4];
    Yr[10] = cm*Y[10] + sm*Y[14];
    Yr[14] = cm*Y[14] - sm*Y[10];

    newcm = c*cm - s*sm;
    sm = c*sm + s*cm;
    cm = newcm;

    // m=3
    Yr[ 9] = cm*Y[ 9] + sm*Y[15];
    Yr[15] = cm*Y[15] - sm*Y[ 9];
}

// routine generated programmatically for evaluating SH basis for degree 3
// inputs (x,y,z) are a point on the sphere (i.e., must be unit length)
// output is vector b with SH basis evaluated at (x,y,z).
//
//void SH_eval_basis_4(in float x, in float y, in float z, inout float b[16]) 
void SH_eval_basis_4(in float x, in float y, in float z, inout float b[16]) 
{
    float z2 = z*z;
    /* m=0 */

    // l=0
    float p_0_0 = CONSTANT(0.282094791773878140);
    b[  0] = p_0_0; // l=0,m=0
    // l=1
    float p_1_0 = CONSTANT(0.488602511902919920)*z;
    b[  2] = p_1_0; // l=1,m=0
    // l=2
    float p_2_0 = CONSTANT(0.946174695757560080)*z2 + CONSTANT(-0.315391565252520050);
    b[  6] = p_2_0; // l=2,m=0
    // l=3
    float p_3_0 = z*(CONSTANT(1.865881662950577000)*z2 + CONSTANT(-1.119528997770346200));
    b[ 12] = p_3_0; // l=3,m=0


    /* m=1 */

    float s_1 = y;
    float c_1 = x;

    // l=1
    float p_1_1 = CONSTANT(-0.488602511902919920);
    b[  1] = p_1_1*s_1; // l=1,m=-1
    b[  3] = p_1_1*c_1; // l=1,m=+1
    // l=2
    float p_2_1 = CONSTANT(-1.092548430592079200)*z;
    b[  5] = p_2_1*s_1; // l=2,m=-1
    b[  7] = p_2_1*c_1; // l=2,m=+1
    // l=3
    float p_3_1 = CONSTANT(-2.285228997322328800)*z2 + CONSTANT(0.457045799464465770);
    b[ 11] = p_3_1*s_1; // l=3,m=-1
    b[ 13] = p_3_1*c_1; // l=3,m=+1


    /* m=2 */

    float s_2 = x*s_1 + y*c_1;
    float c_2 = x*c_1 - y*s_1;

    // l=2
    float p_2_2 = CONSTANT(0.546274215296039590);
    b[  4] = p_2_2*s_2; // l=2,m=-2
    b[  8] = p_2_2*c_2; // l=2,m=+2
    // l=3
    float p_3_2 = CONSTANT(1.445305721320277100)*z;
    b[ 10] = p_3_2*s_2; // l=3,m=-2
    b[ 14] = p_3_2*c_2; // l=3,m=+2


    /* m=3 */

    float s_3 = x*s_2 + y*c_2;
    float c_3 = x*c_2 - y*s_2;

    // l=3
    const float p_3_3 = CONSTANT(-0.590043589926643520);
    b[  9] = p_3_3*s_3; // l=3,m=-3
    b[ 15] = p_3_3*c_3; // l=3,m=+3
}

//
// Rotate ZH (zonal harmonic, function circularly symmetric about z axis) around z axis
//  output is placed in y
//
void SH_zh_rot_4(in vec3 a, in float zh[4], inout float y[16])
{
	//float M_PI = fastAtan(1.0)*4.0;
	float M_PI_LOCAL = atan(1.0)*4.0;
	float sqrtA = sqrt(4*M_PI_LOCAL/(2*0+1));
	float sqrtB = sqrt(4*M_PI_LOCAL/(2*1+1));
	float sqrtC = sqrt(4*M_PI_LOCAL/(2*2+1));
	float sqrtD = sqrt(4*M_PI_LOCAL/(2*3+1));

	float L[4] = float[4](sqrtA, sqrtB, sqrtC, sqrtD);  // L[i] = sqrt(4*M_PI/(2*i+1));

	//return vec4(a.x,a.y,a.z, 1.0);

	//orig//
	SH_eval_basis_4(a.x,a.y,a.z,y);
	//
	//SH_eval_basis_4(a.x,a.z,a.y,y);

	//return vec4(y[0],y[1],y[2], 1.0);

	float F0 = zh[0]*L[0];
	float F1 = zh[1]*L[1];
	float F2 = zh[2]*L[2];
	float F3 = zh[3]*L[3];

	y[ 0] *= F0;
	y[ 1] *= F1;
	y[ 2] *= F1;
	y[ 3] *= F1;
	y[ 4] *= F2;
	y[ 5] *= F2;
	y[ 6] *= F2;
	y[ 7] *= F2;
	y[ 8] *= F2;
	y[ 9] *= F3;
	y[10] *= F3;
	y[11] *= F3;
	y[12] *= F3;
	y[13] *= F3;
	y[14] *= F3;
	y[15] *= F3;
}

//
// routine generated programatically
// Multiply SH vectors f and g resulting in output SH vector y.
//
void SH_product_4(in float f[16], in float g[16], inout float y[16]) {
    float tf,tg,t;
    // [0,0]: 0,
    y[0] = CONSTANT(0.282094792935999980)*f[0]*g[0];

    // [1,1]: 0,6,8,
    tf = CONSTANT(0.282094791773000010)*f[0]+CONSTANT(-0.126156626101000010)*f[6]+CONSTANT(-0.218509686119999990)*f[8];
    tg = CONSTANT(0.282094791773000010)*g[0]+CONSTANT(-0.126156626101000010)*g[6]+CONSTANT(-0.218509686119999990)*g[8];
    y[1]  = tf*g[1]+tg*f[1];
    t = f[1]*g[1];
    y[0] += CONSTANT(0.282094791773000010)*t;
    y[6]  = CONSTANT(-0.126156626101000010)*t;
    y[8]  = CONSTANT(-0.218509686119999990)*t;

    // [1,4]: 3,13,15,
    tf = CONSTANT(0.218509686114999990)*f[3]+CONSTANT(-0.058399170082300000)*f[13]+CONSTANT(-0.226179013157999990)*f[15];
    tg = CONSTANT(0.218509686114999990)*g[3]+CONSTANT(-0.058399170082300000)*g[13]+CONSTANT(-0.226179013157999990)*g[15];
    y[1] += tf*g[4]+tg*f[4];
    y[4]  = tf*g[1]+tg*f[1];
    t = f[1]*g[4]+f[4]*g[1];
    y[3]  = CONSTANT(0.218509686114999990)*t;
    y[13]  = CONSTANT(-0.058399170082300000)*t;
    y[15]  = CONSTANT(-0.226179013157999990)*t;

    // [1,5]: 2,12,14,
    tf = CONSTANT(0.218509686118000010)*f[2]+CONSTANT(-0.143048168103000000)*f[12]+CONSTANT(-0.184674390923000000)*f[14];
    tg = CONSTANT(0.218509686118000010)*g[2]+CONSTANT(-0.143048168103000000)*g[12]+CONSTANT(-0.184674390923000000)*g[14];
    y[1] += tf*g[5]+tg*f[5];
    y[5]  = tf*g[1]+tg*f[1];
    t = f[1]*g[5]+f[5]*g[1];
    y[2]  = CONSTANT(0.218509686118000010)*t;
    y[12]  = CONSTANT(-0.143048168103000000)*t;
    y[14]  = CONSTANT(-0.184674390923000000)*t;

    // [1,6]: 11,
    tf = CONSTANT(0.202300659402999990)*f[11];
    tg = CONSTANT(0.202300659402999990)*g[11];
    y[1] += tf*g[6]+tg*f[6];
    y[6] += tf*g[1]+tg*f[1];
    t = f[1]*g[6]+f[6]*g[1];
    y[11]  = CONSTANT(0.202300659402999990)*t;

    // [1,8]: 9,11,
    tf = CONSTANT(0.226179013155000000)*f[9]+CONSTANT(0.058399170081799998)*f[11];
    tg = CONSTANT(0.226179013155000000)*g[9]+CONSTANT(0.058399170081799998)*g[11];
    y[1] += tf*g[8]+tg*f[8];
    y[8] += tf*g[1]+tg*f[1];
    t = f[1]*g[8]+f[8]*g[1];
    y[9]  = CONSTANT(0.226179013155000000)*t;
    y[11] += CONSTANT(0.058399170081799998)*t;

    // [2,2]: 0,6,
    tf = CONSTANT(0.282094795249000000)*f[0]+CONSTANT(0.252313259986999990)*f[6];
    tg = CONSTANT(0.282094795249000000)*g[0]+CONSTANT(0.252313259986999990)*g[6];
    y[2] += tf*g[2]+tg*f[2];
    t = f[2]*g[2];
    y[0] += CONSTANT(0.282094795249000000)*t;
    y[6] += CONSTANT(0.252313259986999990)*t;

    // [2,6]: 12,
    tf = CONSTANT(0.247766706973999990)*f[12];
    tg = CONSTANT(0.247766706973999990)*g[12];
    y[2] += tf*g[6]+tg*f[6];
    y[6] += tf*g[2]+tg*f[2];
    t = f[2]*g[6]+f[6]*g[2];
    y[12] += CONSTANT(0.247766706973999990)*t;

    // [3,3]: 0,6,8,
    tf = CONSTANT(0.282094791773000010)*f[0]+CONSTANT(-0.126156626101000010)*f[6]+CONSTANT(0.218509686119999990)*f[8];
    tg = CONSTANT(0.282094791773000010)*g[0]+CONSTANT(-0.126156626101000010)*g[6]+CONSTANT(0.218509686119999990)*g[8];
    y[3] += tf*g[3]+tg*f[3];
    t = f[3]*g[3];
    y[0] += CONSTANT(0.282094791773000010)*t;
    y[6] += CONSTANT(-0.126156626101000010)*t;
    y[8] += CONSTANT(0.218509686119999990)*t;

    // [3,6]: 13,
    tf = CONSTANT(0.202300659402999990)*f[13];
    tg = CONSTANT(0.202300659402999990)*g[13];
    y[3] += tf*g[6]+tg*f[6];
    y[6] += tf*g[3]+tg*f[3];
    t = f[3]*g[6]+f[6]*g[3];
    y[13] += CONSTANT(0.202300659402999990)*t;

    // [3,7]: 2,12,14,
    tf = CONSTANT(0.218509686118000010)*f[2]+CONSTANT(-0.143048168103000000)*f[12]+CONSTANT(0.184674390923000000)*f[14];
    tg = CONSTANT(0.218509686118000010)*g[2]+CONSTANT(-0.143048168103000000)*g[12]+CONSTANT(0.184674390923000000)*g[14];
    y[3] += tf*g[7]+tg*f[7];
    y[7]  = tf*g[3]+tg*f[3];
    t = f[3]*g[7]+f[7]*g[3];
    y[2] += CONSTANT(0.218509686118000010)*t;
    y[12] += CONSTANT(-0.143048168103000000)*t;
    y[14] += CONSTANT(0.184674390923000000)*t;

    // [3,8]: 13,15,
    tf = CONSTANT(-0.058399170081799998)*f[13]+CONSTANT(0.226179013155000000)*f[15];
    tg = CONSTANT(-0.058399170081799998)*g[13]+CONSTANT(0.226179013155000000)*g[15];
    y[3] += tf*g[8]+tg*f[8];
    y[8] += tf*g[3]+tg*f[3];
    t = f[3]*g[8]+f[8]*g[3];
    y[13] += CONSTANT(-0.058399170081799998)*t;
    y[15] += CONSTANT(0.226179013155000000)*t;

    // [4,4]: 0,6,
    tf = CONSTANT(0.282094791770000020)*f[0]+CONSTANT(-0.180223751576000010)*f[6];
    tg = CONSTANT(0.282094791770000020)*g[0]+CONSTANT(-0.180223751576000010)*g[6];
    y[4] += tf*g[4]+tg*f[4];
    t = f[4]*g[4];
    y[0] += CONSTANT(0.282094791770000020)*t;
    y[6] += CONSTANT(-0.180223751576000010)*t;

    // [4,5]: 7,
    tf = CONSTANT(0.156078347226000000)*f[7];
    tg = CONSTANT(0.156078347226000000)*g[7];
    y[4] += tf*g[5]+tg*f[5];
    y[5] += tf*g[4]+tg*f[4];
    t = f[4]*g[5]+f[5]*g[4];
    y[7] += CONSTANT(0.156078347226000000)*t;

    // [4,9]: 3,13,
    tf = CONSTANT(0.226179013157999990)*f[3]+CONSTANT(-0.094031597258400004)*f[13];
    tg = CONSTANT(0.226179013157999990)*g[3]+CONSTANT(-0.094031597258400004)*g[13];
    y[4] += tf*g[9]+tg*f[9];
    y[9] += tf*g[4]+tg*f[4];
    t = f[4]*g[9]+f[9]*g[4];
    y[3] += CONSTANT(0.226179013157999990)*t;
    y[13] += CONSTANT(-0.094031597258400004)*t;

    // [4,10]: 2,12,
    tf = CONSTANT(0.184674390919999990)*f[2]+CONSTANT(-0.188063194517999990)*f[12];
    tg = CONSTANT(0.184674390919999990)*g[2]+CONSTANT(-0.188063194517999990)*g[12];
    y[4] += tf*g[10]+tg*f[10];
    y[10]  = tf*g[4]+tg*f[4];
    t = f[4]*g[10]+f[10]*g[4];
    y[2] += CONSTANT(0.184674390919999990)*t;
    y[12] += CONSTANT(-0.188063194517999990)*t;

    // [4,11]: 3,13,15,
    tf = CONSTANT(-0.058399170082300000)*f[3]+CONSTANT(0.145673124078000010)*f[13]+CONSTANT(0.094031597258400004)*f[15];
    tg = CONSTANT(-0.058399170082300000)*g[3]+CONSTANT(0.145673124078000010)*g[13]+CONSTANT(0.094031597258400004)*g[15];
    y[4] += tf*g[11]+tg*f[11];
    y[11] += tf*g[4]+tg*f[4];
    t = f[4]*g[11]+f[11]*g[4];
    y[3] += CONSTANT(-0.058399170082300000)*t;
    y[13] += CONSTANT(0.145673124078000010)*t;
    y[15] += CONSTANT(0.094031597258400004)*t;

    // [5,5]: 0,6,8,
    tf = CONSTANT(0.282094791773999990)*f[0]+CONSTANT(0.090111875786499998)*f[6]+CONSTANT(-0.156078347227999990)*f[8];
    tg = CONSTANT(0.282094791773999990)*g[0]+CONSTANT(0.090111875786499998)*g[6]+CONSTANT(-0.156078347227999990)*g[8];
    y[5] += tf*g[5]+tg*f[5];
    t = f[5]*g[5];
    y[0] += CONSTANT(0.282094791773999990)*t;
    y[6] += CONSTANT(0.090111875786499998)*t;
    y[8] += CONSTANT(-0.156078347227999990)*t;

    // [5,9]: 14,
    tf = CONSTANT(0.148677009677999990)*f[14];
    tg = CONSTANT(0.148677009677999990)*g[14];
    y[5] += tf*g[9]+tg*f[9];
    y[9] += tf*g[5]+tg*f[5];
    t = f[5]*g[9]+f[9]*g[5];
    y[14] += CONSTANT(0.148677009677999990)*t;

    // [5,10]: 3,13,15,
    tf = CONSTANT(0.184674390919999990)*f[3]+CONSTANT(0.115164716490000000)*f[13]+CONSTANT(-0.148677009678999990)*f[15];
    tg = CONSTANT(0.184674390919999990)*g[3]+CONSTANT(0.115164716490000000)*g[13]+CONSTANT(-0.148677009678999990)*g[15];
    y[5] += tf*g[10]+tg*f[10];
    y[10] += tf*g[5]+tg*f[5];
    t = f[5]*g[10]+f[10]*g[5];
    y[3] += CONSTANT(0.184674390919999990)*t;
    y[13] += CONSTANT(0.115164716490000000)*t;
    y[15] += CONSTANT(-0.148677009678999990)*t;

    // [5,11]: 2,12,14,
    tf = CONSTANT(0.233596680327000010)*f[2]+CONSTANT(0.059470803871800003)*f[12]+CONSTANT(-0.115164716491000000)*f[14];
    tg = CONSTANT(0.233596680327000010)*g[2]+CONSTANT(0.059470803871800003)*g[12]+CONSTANT(-0.115164716491000000)*g[14];
    y[5] += tf*g[11]+tg*f[11];
    y[11] += tf*g[5]+tg*f[5];
    t = f[5]*g[11]+f[11]*g[5];
    y[2] += CONSTANT(0.233596680327000010)*t;
    y[12] += CONSTANT(0.059470803871800003)*t;
    y[14] += CONSTANT(-0.115164716491000000)*t;

    // [6,6]: 0,6,
    tf = CONSTANT(0.282094797560000000)*f[0];
    tg = CONSTANT(0.282094797560000000)*g[0];
    y[6] += tf*g[6]+tg*f[6];
    t = f[6]*g[6];
    y[0] += CONSTANT(0.282094797560000000)*t;
    y[6] += CONSTANT(0.180223764527000010)*t;

    // [7,7]: 0,6,8,
    tf = CONSTANT(0.282094791773999990)*f[0]+CONSTANT(0.090111875786499998)*f[6]+CONSTANT(0.156078347227999990)*f[8];
    tg = CONSTANT(0.282094791773999990)*g[0]+CONSTANT(0.090111875786499998)*g[6]+CONSTANT(0.156078347227999990)*g[8];
    y[7] += tf*g[7]+tg*f[7];
    t = f[7]*g[7];
    y[0] += CONSTANT(0.282094791773999990)*t;
    y[6] += CONSTANT(0.090111875786499998)*t;
    y[8] += CONSTANT(0.156078347227999990)*t;

    // [7,10]: 1,9,11,
    tf = CONSTANT(0.184674390919999990)*f[1]+CONSTANT(0.148677009678999990)*f[9]+CONSTANT(0.115164716490000000)*f[11];
    tg = CONSTANT(0.184674390919999990)*g[1]+CONSTANT(0.148677009678999990)*g[9]+CONSTANT(0.115164716490000000)*g[11];
    y[7] += tf*g[10]+tg*f[10];
    y[10] += tf*g[7]+tg*f[7];
    t = f[7]*g[10]+f[10]*g[7];
    y[1] += CONSTANT(0.184674390919999990)*t;
    y[9] += CONSTANT(0.148677009678999990)*t;
    y[11] += CONSTANT(0.115164716490000000)*t;

    // [7,13]: 2,12,14,
    tf = CONSTANT(0.233596680327000010)*f[2]+CONSTANT(0.059470803871800003)*f[12]+CONSTANT(0.115164716491000000)*f[14];
    tg = CONSTANT(0.233596680327000010)*g[2]+CONSTANT(0.059470803871800003)*g[12]+CONSTANT(0.115164716491000000)*g[14];
    y[7] += tf*g[13]+tg*f[13];
    y[13] += tf*g[7]+tg*f[7];
    t = f[7]*g[13]+f[13]*g[7];
    y[2] += CONSTANT(0.233596680327000010)*t;
    y[12] += CONSTANT(0.059470803871800003)*t;
    y[14] += CONSTANT(0.115164716491000000)*t;

    // [7,14]: 15,
    tf = CONSTANT(0.148677009677999990)*f[15];
    tg = CONSTANT(0.148677009677999990)*g[15];
    y[7] += tf*g[14]+tg*f[14];
    y[14] += tf*g[7]+tg*f[7];
    t = f[7]*g[14]+f[14]*g[7];
    y[15] += CONSTANT(0.148677009677999990)*t;

    // [8,8]: 0,6,
    tf = CONSTANT(0.282094791770000020)*f[0]+CONSTANT(-0.180223751576000010)*f[6];
    tg = CONSTANT(0.282094791770000020)*g[0]+CONSTANT(-0.180223751576000010)*g[6];
    y[8] += tf*g[8]+tg*f[8];
    t = f[8]*g[8];
    y[0] += CONSTANT(0.282094791770000020)*t;
    y[6] += CONSTANT(-0.180223751576000010)*t;

    // [8,9]: 11,
    tf = CONSTANT(-0.094031597259499999)*f[11];
    tg = CONSTANT(-0.094031597259499999)*g[11];
    y[8] += tf*g[9]+tg*f[9];
    y[9] += tf*g[8]+tg*f[8];
    t = f[8]*g[9]+f[9]*g[8];
    y[11] += CONSTANT(-0.094031597259499999)*t;

    // [8,13]: 15,
    tf = CONSTANT(-0.094031597259499999)*f[15];
    tg = CONSTANT(-0.094031597259499999)*g[15];
    y[8] += tf*g[13]+tg*f[13];
    y[13] += tf*g[8]+tg*f[8];
    t = f[8]*g[13]+f[13]*g[8];
    y[15] += CONSTANT(-0.094031597259499999)*t;

    // [8,14]: 2,12,
    tf = CONSTANT(0.184674390919999990)*f[2]+CONSTANT(-0.188063194517999990)*f[12];
    tg = CONSTANT(0.184674390919999990)*g[2]+CONSTANT(-0.188063194517999990)*g[12];
    y[8] += tf*g[14]+tg*f[14];
    y[14] += tf*g[8]+tg*f[8];
    t = f[8]*g[14]+f[14]*g[8];
    y[2] += CONSTANT(0.184674390919999990)*t;
    y[12] += CONSTANT(-0.188063194517999990)*t;

    // [9,9]: 0,6,
    tf = CONSTANT(0.282094791766999970)*f[0]+CONSTANT(-0.210261043508000010)*f[6];
    tg = CONSTANT(0.282094791766999970)*g[0]+CONSTANT(-0.210261043508000010)*g[6];
    y[9] += tf*g[9]+tg*f[9];
    t = f[9]*g[9];
    y[0] += CONSTANT(0.282094791766999970)*t;
    y[6] += CONSTANT(-0.210261043508000010)*t;

    // [10,10]: 0,
    tf = CONSTANT(0.282094791771999980)*f[0];
    tg = CONSTANT(0.282094791771999980)*g[0];
    y[10] += tf*g[10]+tg*f[10];
    t = f[10]*g[10];
    y[0] += CONSTANT(0.282094791771999980)*t;

    // [11,11]: 0,6,8,
    tf = CONSTANT(0.282094791773999990)*f[0]+CONSTANT(0.126156626101000010)*f[6]+CONSTANT(-0.145673124078999990)*f[8];
    tg = CONSTANT(0.282094791773999990)*g[0]+CONSTANT(0.126156626101000010)*g[6]+CONSTANT(-0.145673124078999990)*g[8];
    y[11] += tf*g[11]+tg*f[11];
    t = f[11]*g[11];
    y[0] += CONSTANT(0.282094791773999990)*t;
    y[6] += CONSTANT(0.126156626101000010)*t;
    y[8] += CONSTANT(-0.145673124078999990)*t;

    // [12,12]: 0,6,
    tf = CONSTANT(0.282094799871999980)*f[0]+CONSTANT(0.168208852954000010)*f[6];
    tg = CONSTANT(0.282094799871999980)*g[0]+CONSTANT(0.168208852954000010)*g[6];
    y[12] += tf*g[12]+tg*f[12];
    t = f[12]*g[12];
    y[0] += CONSTANT(0.282094799871999980)*t;
    y[6] += CONSTANT(0.168208852954000010)*t;

    // [13,13]: 0,6,8,
    tf = CONSTANT(0.282094791773999990)*f[0]+CONSTANT(0.126156626101000010)*f[6]+CONSTANT(0.145673124078999990)*f[8];
    tg = CONSTANT(0.282094791773999990)*g[0]+CONSTANT(0.126156626101000010)*g[6]+CONSTANT(0.145673124078999990)*g[8];
    y[13] += tf*g[13]+tg*f[13];
    t = f[13]*g[13];
    y[0] += CONSTANT(0.282094791773999990)*t;
    y[6] += CONSTANT(0.126156626101000010)*t;
    y[8] += CONSTANT(0.145673124078999990)*t;

    // [14,14]: 0,
    tf = CONSTANT(0.282094791771999980)*f[0];
    tg = CONSTANT(0.282094791771999980)*g[0];
    y[14] += tf*g[14]+tg*f[14];
    t = f[14]*g[14];
    y[0] += CONSTANT(0.282094791771999980)*t;

    // [15,15]: 0,6,
    tf = CONSTANT(0.282094791766999970)*f[0]+CONSTANT(-0.210261043508000010)*f[6];
    tg = CONSTANT(0.282094791766999970)*g[0]+CONSTANT(-0.210261043508000010)*g[6];
    y[15] += tf*g[15]+tg*f[15];
    t = f[15]*g[15];
    y[0] += CONSTANT(0.282094791766999970)*t;
    y[6] += CONSTANT(-0.210261043508000010)*t;
}


float hightFieldPyramidLowToHigh(float tex_x, float tex_y, int i, int nlevels )
{
	int level = i*2+1;
	int N = nlevels*2;
	return texture(heightField, vec3(tex_x, tex_y ,  float(N-level))).z;
}

//Construct blocking angle pyramid omega_i from height pyramid f_i
vec4 ShadeEnvLight(vec2 tex, int numAzimuthalSamples, int levelOffset)
{
    float swathSize = (2.0 * M_PI) / numAzimuthalSamples;
    const float l = 1.0; // Multi-res step size
   
    float f_x0[NUM_PYRAMID_LEVELS];
    float multiresUVStepSizes[NUM_PYRAMID_LEVELS];
    
    for( int i = 0; i < NUM_PYRAMID_LEVELS; i++ ) 
	{
        f_x0[i] = hightFieldPyramidLowToHigh( tex.x, tex.y, i, NUM_PYRAMID_LEVELS );
		multiresUVStepSizes[i] = pow( multiResBase, -i );
    }

	//return vec4(multiresUVStepSizes[0],multiresUVStepSizes[1],multiresUVStepSizes[2],1.0);

	// Iterate over all azimuthal swaths / visibility
    float D[NUM_PYRAMID_LEVELS];
    float maxBlockingAngles[NUM_ENV_SWATHS];   
    float currentPhi0s[NUM_ENV_SWATHS];
	vec2 f_delta[NUM_PYRAMID_LEVELS];

	vec2 cosSinCurrentPhi0s[NUM_ENV_SWATHS];

    for(int iSwathStart = 0; iSwathStart < numAzimuthalSamples; iSwathStart++) 
	{
        currentPhi0s[iSwathStart] = iSwathStart * swathSize;
        cosSinCurrentPhi0s[iSwathStart] = vec2( cos( currentPhi0s[iSwathStart] ), sin( currentPhi0s[iSwathStart] ) );
        
        // Calculate D_i and w_i for this swath
        //maxBlockingAngles[iSwathStart] = 0;
        //for(int i = 0; i < NUM_PYRAMID_LEVELS; i++) 
		for(int i = NUM_PYRAMID_LEVELS-1 ; i >= 0 ; i--)
		{
            int level = min(i + levelOffset, NUM_PYRAMID_LEVELS - 1);
            f_delta[i] = multiresUVStepSizes[i] * cosSinCurrentPhi0s[iSwathStart];
            float f_x1 = hightFieldPyramidLowToHigh(tex.x + l * f_delta[i].x, tex.y + l * f_delta[i].y, level, NUM_PYRAMID_LEVELS);
            D[i] = (f_x1 - f_x0[level]) / (l * multiresUVStepSizes[i]);

            // No 1D B-Spline interpolation (environmental lighting)
			//// Just use sample points and can take max over D instead of w and perform 1 atan!
            maxBlockingAngles[iSwathStart] = max( maxBlockingAngles[iSwathStart], D[i] );
		}

		//return vec4(maxBlockingAngles[4],maxBlockingAngles[5],maxBlockingAngles[6],1.0);
		
        //maxBlockingAngles[iSwathStart] = fastAtan(maxBlockingAngles[iSwathStart]);
        maxBlockingAngles[iSwathStart] = atan(maxBlockingAngles[iSwathStart]);
	}

	//return vec4(maxBlockingAngles[1],maxBlockingAngles[2],maxBlockingAngles[3],0.0);

	float visVector[16];/* = float[16](0.0, 0.0, 0.0, 0.0, 
									0.0, 0.0, 0.0, 0.0, 
									0.0, 0.0, 0.0, 0.0, 
									0.0, 0.0, 0.0, 0.0);*/
	float rotatedVisSwath[16] = float[16](0.0, 0.0, 0.0, 0.0, 
										  0.0, 0.0, 0.0, 0.0, 
										  0.0, 0.0, 0.0, 0.0, 
										  0.0, 0.0, 0.0, 0.0);
	// This will store the visibility and transfer vector (used to shade) for this texel
	float texelSHVector[16] = float[16](0.0, 0.0, 0.0, 0.0, 
										0.0, 0.0, 0.0, 0.0, 
										0.0, 0.0, 0.0, 0.0, 
										0.0, 0.0, 0.0, 0.0);

	for( int visWedge = 0; visWedge < numAzimuthalSamples - 1; visWedge++ ) 
	{
        vec2 blockingAngleUVs = vec2( maxBlockingAngles[visWedge] * ONE_OVER_HALF_PI, maxBlockingAngles[visWedge+1] * ONE_OVER_HALF_PI );

		vec4 vis1local = texture2D( vis4, blockingAngleUVs);
		visVector[0] = vis1local.x;
		visVector[1] = vis1local.y;
		visVector[2] = vis1local.z;
		visVector[3] = vis1local.w;

		vec4 vis2 = texture2D( vis8, blockingAngleUVs );
		visVector[4] = vis2.x;
		visVector[5] = vis2.y;
		visVector[6] = vis2.z;
		visVector[7] = vis2.w;

		vec4 vis3 = texture2D( vis12, blockingAngleUVs );
		visVector[8] = vis3.x;
		visVector[9] = vis3.y;
		visVector[10] = vis3.z;
		visVector[11] = vis3.w;

		vec4 vis5 = texture2D( vis16, blockingAngleUVs );
		visVector[12] = vis5.x;
		visVector[13] = vis5.y;
		visVector[14] = vis5.z;
		visVector[15] = vis5.w;


		// Given SH Vis vector, get transfer vector and use to shade with lighting vector
        SH_rotatez_4( cosSinCurrentPhi0s[visWedge].x, cosSinCurrentPhi0s[visWedge].y, visVector, rotatedVisSwath);

		for( int a = 0; a < 16; a++ ) 
		{
            texelSHVector[a] += rotatedVisSwath[a];
        }
	}
	
	// special case last wedge
	{
        vec2 blockingAngleUVs = vec2( maxBlockingAngles[numAzimuthalSamples - 1] * ONE_OVER_HALF_PI, maxBlockingAngles[0] * ONE_OVER_HALF_PI );

		vec4 vis1local = texture2D( vis4, blockingAngleUVs);
		visVector[0] = vis1local.x;
		visVector[1] = vis1local.y;
		visVector[2] = vis1local.z;
		visVector[3] = vis1local.w;

		vec4 vis2 = texture2D( vis8, blockingAngleUVs );
		visVector[4] = vis2.x;
		visVector[5] = vis2.y;
		visVector[6] = vis2.z;
		visVector[7] = vis2.w;

		vec4 vis3 = texture2D( vis12, blockingAngleUVs );
		visVector[8] = vis3.x;
		visVector[9] = vis3.y;
		visVector[10] = vis3.z;
		visVector[11] = vis3.w;

		vec4 vis5 = texture2D( vis16, blockingAngleUVs );
		visVector[12] = vis5.x;
		visVector[13] = vis5.y;
		visVector[14] = vis5.z;
		visVector[15] = vis5.w;


		// Given SH Vis vector, get transfer vector and use to shade with lighting vector
        SH_rotatez_4( cosSinCurrentPhi0s[numAzimuthalSamples - 1].x, cosSinCurrentPhi0s[numAzimuthalSamples - 1].y, visVector, rotatedVisSwath);

		for( int a = 0; a < 16; a++ ) 
		{
            texelSHVector[a] += rotatedVisSwath[a];
        }
	}

	// Convert from visibility to transfer
    float rotated_hemi[16]  = float[16](0.0, 0.0, 0.0, 0.0, 
										0.0, 0.0, 0.0, 0.0, 
										0.0, 0.0, 0.0, 0.0, 
										0.0, 0.0, 0.0, 0.0);

	float transferSH[16] = float[16](0.0, 0.0, 0.0, 0.0, 
									 0.0, 0.0, 0.0, 0.0, 
									 0.0, 0.0, 0.0, 0.0, 
									 0.0, 0.0, 0.0, 0.0);

	//return vec4(texture( test, vec2( tex.x, tex.y )).xyz, 1.0);

	SH_zh_rot_4( texture( test, vec2( tex.x, tex.y )).xyz, zhemi_canon, rotated_hemi );
    SH_product_4( rotated_hemi, texelSHVector, transferSH );

	vec4 transfer[4];
	transfer[0] = vec4(transferSH[0],transferSH[1],transferSH[2],transferSH[3]);
	transfer[1] = vec4(transferSH[4],transferSH[5],transferSH[6],transferSH[7]);
	transfer[2] = vec4(transferSH[8],transferSH[9],transferSH[10],transferSH[11]);
	transfer[3] = vec4(transferSH[12],transferSH[13],transferSH[14],transferSH[15]);

	vec4 RLight[4];
	RLight[0] = vec4(RedLightEnv[0], RedLightEnv[1], RedLightEnv[2], RedLightEnv[3]);
	RLight[1] = vec4(RedLightEnv[4], RedLightEnv[5], RedLightEnv[6], RedLightEnv[7]);
	RLight[2] = vec4(RedLightEnv[8], RedLightEnv[9], RedLightEnv[10], RedLightEnv[11]);
	RLight[3] = vec4(RedLightEnv[12], RedLightEnv[13], RedLightEnv[14], RedLightEnv[15]);

	vec4 GLight[4];
	GLight[0] = vec4(GreenLightEnv[0], GreenLightEnv[1], GreenLightEnv[2], GreenLightEnv[3]);
	GLight[1] = vec4(GreenLightEnv[4], GreenLightEnv[5], GreenLightEnv[6], GreenLightEnv[7]);
	GLight[2] = vec4(GreenLightEnv[8], GreenLightEnv[9], GreenLightEnv[10], GreenLightEnv[11]);
	GLight[3] = vec4(GreenLightEnv[12], GreenLightEnv[13], GreenLightEnv[14], GreenLightEnv[15]);

	vec4 BLight[4];
	BLight[0] = vec4(BlueLightEnv[0], BlueLightEnv[1], BlueLightEnv[2], BlueLightEnv[3]);
	BLight[1] = vec4(BlueLightEnv[4], BlueLightEnv[5], BlueLightEnv[6], BlueLightEnv[7]);
	BLight[2] = vec4(BlueLightEnv[8], BlueLightEnv[9], BlueLightEnv[10], BlueLightEnv[11]);
	BLight[3] = vec4(BlueLightEnv[12], BlueLightEnv[13], BlueLightEnv[14], BlueLightEnv[15]);

	//return vec4(BLight[3].xyz,1.0);
	// transfer dot light
    vec4 Color;
    Color.r = dot( transfer[0], RLight[0] ) + 
			  dot( transfer[1], RLight[1] ) + 
			  dot( transfer[2], RLight[2] ) + 
              dot( transfer[3], RLight[3] );
    Color.g = dot( transfer[0], GLight[0] ) + 
			  dot( transfer[1], GLight[1] ) + 
			  dot( transfer[2], GLight[2] ) + 
              dot( transfer[3], GLight[3] );
    Color.b = dot( transfer[0], BLight[0] ) + 
			  dot( transfer[1], BLight[1] ) + 
			  dot( transfer[2], BLight[2] ) + 
              dot( transfer[3], BLight[3] );
	   
	Color *= 1.0 / M_PI;
    Color.a = 1.0;

	//return vec4(1.0,0.0,0.0, 1.0);

    return Color;
}

void main()
{
	//TODO
	ex_Color = ShadeEnvLight(ex_TexCoord, NUM_ENV_SWATHS, offset);
}