//-------------------------------------------------------------------------------
//
//Tangent space normal mapping shader for a single directional light source.
//
//The tangent vector is passed to the vertex shader in gl_MultiTexCoord1. The
//tangent vector is assumed to be a four component vector. The tangent vector's
//w component indicates the handedness of the local tangent space at this vertex.
//The handedness is used to calculate the bitangent vector. The reason for the
//inclusion of the handedness component is to allow for triangles with mirrored
//texture mappings.
//
//-------------------------------------------------------------------------------

#version 140

uniform sampler2D ColorMap;
uniform sampler2D NormalMap;

uniform float MaterialAlpha;
uniform float MaterialFrontShininess;
uniform vec4 LightFrontAmbient;
uniform vec4 LightFrontDiffuse;
uniform vec4 LightFrontSpecular;
uniform vec4 LightFrontAmbient;

in vec3 LightDirection;
in vec3 LightHalfVector;
in vec4 SceneColor;

out vec4 FragColor;

void main()
{
    vec3 n = normalize(texture2D(NormalMap, gl_TexCoord[0].st).rgb * 2.0 - 1.0);
    vec3 l = normalize(LightDirection);
    vec3 h = normalize(LightHalfVector);

    float nDotL = max(0.0, dot(n, l));
    float nDotH = max(0.0, dot(n, h));
    float power = (nDotL == 0.0) ? 0.0 : pow(nDotH, MaterialFrontShininess);
    
    vec4 ambient = LightFrontAmbient;
    vec4 diffuse = LightFrontDiffuse * nDotL;
    vec4 specular = LightFrontSpecular * power;
    vec4 color = SceneColor + ambient + diffuse + specular;
    
    FragColor = color * texture2D(ColorMap, gl_TexCoord[0].st);
    FragColor.a = MaterialAlpha;
}
