//-------------------------------------------------------------------------------
//
//Tangent space normal mapping shader for a single directional light source.
//
//The tangent vector is passed to the vertex shader in gl_MultiTexCoord1. The
//tangent vector is assumed to be a four component vector. The tangent vector's
//w component indicates the handedness of the local tangent space at this vertex.
//The handedness is used to calculate the bitangent vector. The reason for the
//inclusion of the handedness component is to allow for triangles with mirrored
//texture mappings.
//
//-------------------------------------------------------------------------------

#version 140

uniform mat4 ViewMatrix;
uniform mat4 ProjectionMatrix;
uniform mat4 NormalMatrix;

uniform vec3 LightSourcePosition;
uniform vec3 LightSourceHalfVector;

in vec3 Vertex;
in vec3 Normal;
in vec4 MultiTexCoord1;
in vec4 MultiTexCoord2;

out vec3 LightDirection;
out vec3 LightHalfVector;

out vec4 TexCoord;

void main()
{
    gl_Position = ProjectionMatrix * ModelViewMatrix * Vertex;
    TexCoord = MultiTexCoord1;

    vec3 n = normalize(NormalMatrix * Normal);
    vec3 t = normalize(NormalMatrix * MultiTexCoord2.xyz);
    vec3 b = cross(n, t) * MultiTexCoord2.w;

    mat3 tbnMatrix = mat3(t.x, b.x, n.x,
                          t.y, b.y, n.y,
                          t.z, b.z, n.z);

    LightDirection = LightSourcePosition;
    LightDirection = tbnMatrix * LightDirection;

    LightHalfVector = LightSourceHalfVector;
    LightHalfVector = tbnMatrix * LightHalfVector;
}
