#version 140

uniform sampler2D DiffuseColorMap;
uniform sampler2D NormalMap;

uniform vec4 LightFrontAmbient;
uniform vec4 LightFrontDiffuse;
uniform vec4 LightFrontSpecular;
uniform float MaterialFrontShininess;
uniform vec4 SceneColor;

in vec2 ex_TexCoord;
in vec3 LightDirection;
in vec3 LightHalfVector;

out vec4 FragColor;

void main()
{
    vec3 n = normalize(texture2D(NormalMap, ex_TexCoord.xy).rgb * 2.0 - 1.0);
    vec3 l = normalize(LightDirection);
    vec3 h = normalize(LightHalfVector);

    float nDotL = max(0.0, dot(n, l));
    float nDotH = max(0.0, dot(n, h));
    float power = (nDotL == 0.0) ? 0.0 : pow(nDotH, MaterialFrontShininess);
    
    vec4 ambient = LightFrontAmbient;
    vec4 diffuse = LightFrontDiffuse * nDotL;
    vec4 specular = LightFrontSpecular * power;
    vec4 color = SceneColor + ambient + diffuse + specular;
    
    FragColor = color * texture2D(DiffuseColorMap, ex_TexCoord.xy);
    FragColor.a = 1.0;
}