#version 140

uniform sampler2D samLinearClamp;
uniform float size;

in vec2 ex_TexCoord;
out vec4 ex_Color;

float texelSize = 1.0 / size;
const float normalStrength = 1.0;//0.5;//8;

vec4 CalculateNormalsOrig(vec2 texCoord)
{
    vec3 normal = vec3(0.0, 0.0, 0.0);
    float delta = 1.0 / float(size);
    
    float thisPointZ = texture2D( samLinearClamp, texCoord ).z;
    vec3 left, right, top, bottom;
        
    // The HF is arbitrarily stored in g_txRenderTexturePyramidLevelCoarse
    left = vec3( -delta, 0.0, texture2D( samLinearClamp, vec2( texCoord.x - delta, texCoord.y ) ).z - thisPointZ );
    right = vec3( delta, 0.0, texture2D( samLinearClamp, vec2( texCoord.x + delta, texCoord.y ) ).z - thisPointZ );
    top = vec3( 0.0, delta, texture2D( samLinearClamp, vec2( texCoord.x, texCoord.y + delta ) ).z - thisPointZ );
    bottom = vec3( 0.0, -delta, texture2D( samLinearClamp, vec2( texCoord.x, texCoord.y - delta) ).z - thisPointZ );
 
    
    left = normalize( left );
    right = normalize( right );
    top = normalize( top );
    bottom = normalize( bottom );
    
    // TODO: optimize the cross products
    normal += normalize( cross( left, bottom ) );
    normal += normalize( cross( bottom, right ) );
    normal += normalize( cross( right, top ) );
    normal += normalize( cross( top, left ) );
    
    ///normal *= 0.25f;
    normal = normalize(normal);

    return vec4( normal, 0.0 ); // Use first 3 components when reading.... can't set format = R32G32B32 to RenderTarget in DX10
}


vec4 CalculateNormals(in vec2 texCoord)
{
    float tl = abs(texture2D (samLinearClamp, texCoord + texelSize * vec2(-1, -1)).x);   // top left
    float  l = abs(texture2D (samLinearClamp, texCoord + texelSize * vec2(-1,  0)).x);   // left
    float bl = abs(texture2D (samLinearClamp, texCoord + texelSize * vec2(-1,  1)).x);   // bottom left
    float  t = abs(texture2D (samLinearClamp, texCoord + texelSize * vec2( 0, -1)).x);   // top
    float  b = abs(texture2D (samLinearClamp, texCoord + texelSize * vec2( 0,  1)).x);   // bottom
    float tr = abs(texture2D (samLinearClamp, texCoord + texelSize * vec2( 1, -1)).x);   // top right
    float  r = abs(texture2D (samLinearClamp, texCoord + texelSize * vec2( 1,  0)).x);   // right
    float br = abs(texture2D (samLinearClamp, texCoord + texelSize * vec2( 1,  1)).x);   // bottom right

    // Compute dx using Sobel:
    //           -1 0 1 
    //           -2 0 2
    //           -1 0 1

    float dX = tr + 2*r + br -tl - 2*l - bl;

    //  Compute dy using Sobel:
    //           -1 -2 -1 
    //            0  0  0
    //            1  2  1

    float dY = bl + 2*b + br -tl - 2*t - tr;
   //  Build the normalized normal

    vec4 N = vec4(normalize(vec3(dX, 1.0 / normalStrength, dY)), 1.0);
    //convert (-1.0 , 1.0) to (0.0 , 1.0), if needed

    return N * 0.5 + 0.5;
}

void main()
{
	ex_Color = CalculateNormals(ex_TexCoord);
	//ex_Color = CalculateNormalsOrig(ex_TexCoord);
	//ex_Color = vec4(1.0, 0.0, 0.0, 1.0);
}